import copy

import numpy as np
import pytest

import qcelemental
from qcelemental.models import Molecule
from qcelemental.testing import compare, compare_molrecs, compare_recursive, compare_values, tnm

_arrays_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_arrays"}
_string_prov_stamp = {"creator": "QCElemental", "version": "1.0", "routine": "qcelemental.molparse.from_string"}

_trans_molrec_to_model = {
    "geometry": "geom",
    "mass_numbers": "elea",
    "atomic_numbers": "elez",
    "symbols": "elem",
    "masses": "mass",
    "real": "real",
    "atom_labels": "elbl",
    "fix_com": "fix_com",
    "fix_orientation": "fix_orientation",
    "fragment_charges": "fragment_charges",
    "fragment_multiplicities": "fragment_multiplicities",
    "molecular_charge": "molecular_charge",
    "molecular_multiplicity": "molecular_multiplicity",
}


def _check_eq_molrec_minimal_model(keepers, model, molrec=None):
    usual_stored = set(
        ["geometry", "symbols", "molecular_charge", "molecular_multiplicity", "fix_com", "fix_orientation"]
    )
    usual_filtered = set(
        [
            "masses",
            "mass_numbers",
            "real",
            "atomic_numbers",
            "atom_labels",
            "connectivity",
            "fragments",
            "fragment_charges",
            "fragment_multiplicities",
        ]
    )

    for field in usual_filtered.difference(keepers):
        assert field not in model, f"Field '{field}' strangely present: {model[field]}"

    for field in usual_stored.union(keepers):
        assert field in model, f"Field '{field}' strangely absent"
        if molrec and field not in ["geometry", "fragments"]:
            skf = model[field]
            mrf = molrec[_trans_molrec_to_model[field]]
            if field in ["molecular_charge", "fragment_charges", "geometry", "masses"]:
                assert compare_values(mrf, skf, field, atol=1.0e-6)  # MASS_NOISE
            else:
                assert compare(mrf, skf, field)  # MASS_NOISE


subject1 = """O 0 0   0
no_com

H 1 ,, 0 \t  0 # stuff-n-nonsense"""

ans1 = {
    "geom": [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
    "elbl": ["O", "H"],
    "fix_com": True,
    "fragment_separators": [],
    "fragment_charges": [None],
    "fragment_multiplicities": [None],
    "fragment_files": [],
    "geom_hints": [],
    "hint_types": [],
}

fullans1a = {
    "geom": np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
    "elea": np.array([16, 1]),
    "elez": np.array([8, 1]),
    "elem": np.array(["O", "H"]),
    "mass": np.array([15.99491462, 1.00782503]),
    "real": np.array([True, True]),
    "elbl": np.array(["", ""]),
    "units": "Angstrom",
    "fix_com": True,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [2],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 2,
}
fullans1c = copy.deepcopy(fullans1a)
fullans1c.update(
    {"fragment_charges": [1.0], "fragment_multiplicities": [1], "molecular_charge": 1.0, "molecular_multiplicity": 1}
)


def test_psi4_qm_1a():
    subject = subject1
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans1, intermed, atol=1.0e-4)
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")

    kmol = Molecule.from_data(subject)
    _check_eq_molrec_minimal_model([], kmol.dict(), fullans)


def test_psi4_qm_1ab():
    subject = subject1
    ans = copy.deepcopy(ans1)
    ans["fix_orientation"] = False
    ans["fix_com"] = False
    fullans = copy.deepcopy(fullans1a)
    fullans["fix_com"] = False
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(
        subject, return_processed=True, fix_orientation=False, fix_com=False
    )
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_psi4_qm_1b():
    subject = "\n" + "\t" + subject1 + "\n\n"
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans1, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_psi4_qm_1c():
    subject = "1 1\n  -- \n" + subject1
    ans = copy.deepcopy(ans1)
    ans.update({"molecular_charge": 1.0, "molecular_multiplicity": 1})
    fullans = copy.deepcopy(fullans1c)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")

    kmol = Molecule.from_data(subject)
    _check_eq_molrec_minimal_model([], kmol.dict(), fullans)


def test_psi4_qm_1d():
    subject = subject1 + "\n1 1"
    ans = copy.deepcopy(ans1)
    ans.update({"fragment_charges": [1.0], "fragment_multiplicities": [1]})
    fullans = copy.deepcopy(fullans1c)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_psi4_qm_1e():
    """duplicate com"""
    subject = subject1 + "\n  nocom"

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_qm_1f():

    qcelemental.molparse.from_arrays(
        geom=np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
        elez=np.array([8, 1]),
        units="Angstrom",
        fix_com=True,
        fix_orientation=False,
    )


def test_psi4_qm_iutautoobig_error_1g():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
            elez=np.array([8, 1]),
            input_units_to_au=1.1 / 0.52917721067,
            units="Angstrom",
            fix_com=True,
            fix_orientation=False,
        )

    assert "No big perturbations to physical constants" in str(e.value)


def test_psi4_qm_iutau_1h():
    fullans = copy.deepcopy(fullans1a)
    iutau = 1.01 / 0.52917721067
    fullans["input_units_to_au"] = iutau
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        geom=np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
        elez=np.array([8, 1]),
        input_units_to_au=iutau,
        units="Angstrom",
        fix_com=True,
        fix_orientation=False,
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")

    smol = qcelemental.molparse.to_string(final, dtype="xyz", units="Bohr")
    rsmol = """2 au
0 2 HO
O                     0.000000000000     0.000000000000     0.000000000000
H                     1.908623386712     0.000000000000     0.000000000000
"""
    assert compare(rsmol, smol, tnm() + ": str")


def test_psi4_qm_iutau_1i():
    fullans = copy.deepcopy(fullans1a)
    iutau = 1.01 / 0.52917721067
    fullans["input_units_to_au"] = iutau
    fullans["fix_symmetry"] = "cs"
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        geom=np.array([0.0, 0.0, 0.0, 1.0, 0.0, 0.0]),
        elez=np.array([8, 1]),
        input_units_to_au=iutau,
        units="Angstrom",
        fix_symmetry="CS",
        fix_com=True,
        fix_orientation=False,
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")

    kmol = qcelemental.molparse.to_schema(final, dtype=1, units="Bohr")
    schema14_1_iutau = {
        "schema_name": "qc_schema_input",
        "schema_version": 1,
        "molecule": {
            "validated": True,
            "geometry": [0.0, 0.0, 0.0, 1.908623386712, 0.0, 0.0],
            "symbols": ["O", "H"],
            "atomic_numbers": [8, 1],
            "mass_numbers": [16, 1],
            "atom_labels": ["", ""],
            "fragments": [[0, 1]],
            "fragment_charges": [0.0],
            "fragment_multiplicities": [2],
            "masses": [15.99491462, 1.00782503],
            "name": "HO",
            "fix_com": True,
            "fix_orientation": False,
            "fix_symmetry": "cs",
            "molecular_charge": 0.0,
            "molecular_multiplicity": 2,
            "real": [True, True],
            "provenance": _arrays_prov_stamp,
        },
    }

    assert compare_molrecs(schema14_1_iutau["molecule"], kmol["molecule"], tnm() + ": sch", atol=1.0e-8)


subject2 = [
    """
6Li 0.0 0.0 0.0
  units  a.u.
H_specIAL@2.014101  100 0 0""",
    """@Ne 2 4 6""",
    """h .0,1,2
Gh(he3) 0 1 3
noreorient""",
]

ans2 = {
    "geom": [0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0],
    "elbl": ["6Li", "H_specIAL@2.014101", "@Ne", "h", "Gh(he3)"],
    "units": "Bohr",
    "fix_orientation": True,
    "fragment_separators": [2, 3],
    "fragment_charges": [None, None, None],
    "fragment_multiplicities": [None, None, None],
    "fragment_files": [],
    "geom_hints": [],
    "hint_types": [],
}

fullans2 = {
    "geom": np.array([0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0]),
    "elea": np.array([6, 2, 20, 1, 4]),
    "elez": np.array([3, 1, 10, 1, 2]),
    "elem": np.array(["Li", "H", "Ne", "H", "He"]),
    "mass": np.array([6.015122794, 2.014101, 19.99244017542, 1.00782503, 4.00260325415]),
    "real": np.array([True, True, False, True, False]),
    "elbl": np.array(["", "_special", "", "", "3"]),
    "units": "Bohr",
    "fix_com": False,
    "fix_orientation": True,
    "fragment_separators": [2, 3],
}
fullans2_unnp = copy.deepcopy(fullans2)
fullans2_unnp.update(
    {
        "geom": [0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0],
        "elea": [6, 2, 20, 1, 4],
        "elez": [3, 1, 10, 1, 2],
        "elem": ["Li", "H", "Ne", "H", "He"],
        "mass": [6.015122794, 2.014101, 19.99244017542, 1.00782503, 4.00260325415],
        "real": [True, True, False, True, False],
        "elbl": ["", "_special", "", "", "3"],
    }
)


def test_psi4_qm_2a():
    subject = "\n--\n".join(subject2)
    fullans = copy.deepcopy(fullans2)
    fullans_unnp = copy.deepcopy(fullans2_unnp)
    ud = {
        "molecular_charge": 0.0,
        "molecular_multiplicity": 2,
        "fragment_charges": [0.0, 0.0, 0.0],
        "fragment_multiplicities": [1, 1, 2],
    }
    fullans.update(ud)
    fullans_unnp.update(ud)
    fullans["provenance"] = _string_prov_stamp
    fullans_unnp["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans2, intermed, tnm() + ": intermediate")
    final_unnp = qcelemental.util.unnp(final["qm"])
    assert compare_molrecs(fullans_unnp, final_unnp, tnm() + ": full unnp")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")

    kmol = Molecule.from_data(subject)
    _check_eq_molrec_minimal_model(
        ["fragments", "fragment_charges", "fragment_multiplicities", "mass_numbers", "masses", "atom_labels", "real"],
        kmol.dict(),
        fullans,
    )


def test_psi4_qm_2b():
    subject = copy.deepcopy(subject2)
    subject.insert(0, "1 3")
    subject = "\n--\n".join(subject)
    ans = copy.deepcopy(ans2)
    ans.update({"molecular_charge": 1.0, "molecular_multiplicity": 3})
    fullans = copy.deepcopy(fullans2)
    fullans.update(
        {
            "molecular_charge": 1.0,
            "molecular_multiplicity": 3,
            "fragment_charges": [1.0, 0.0, 0.0],
            "fragment_multiplicities": [2, 1, 2],
        }
    )
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_psi4_qm_2c():
    """double overall chg/mult spec"""
    subject = copy.deepcopy(subject2)
    subject.insert(0, "1 3\n1 3")
    subject = "\n--\n".join(subject)

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_qm_2d():
    """trailing comma"""
    subject = copy.deepcopy(subject2)
    subject.insert(0, "H 10,10,10,")
    subject = "\n--\n".join(subject)

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


# def test_psi4_qm_2e():
#    """empty fragment"""
#    subject = copy.deepcopy(subject2)
#    subject.insert(2, '\n')
#    subject = '\n--\n'.join(subject)
#
#    with pytest.raises(qcelemental.MoleculeFormatError):
#        final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_qm_2f():
    """double frag chgmult"""
    subject = copy.deepcopy(subject2)
    subject[1] += "\n 1 2\n 5 6"
    subject = "\n--\n".join(subject)

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_qm_2g():
    """illegal chars in nucleus"""
    subject = copy.deepcopy(subject2)
    subject[1] = """@Ne_{CN}_O 2 4 6"""
    subject = "\n--\n".join(subject)

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_qm_3():
    """psi4/psi4#731"""
    subject = """0 1
Mg 0 0"""

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_string(subject, return_processed=True)


subject5 = """
efp C6H6 -0.30448173 -2.24210052 -0.29383131 -0.642499 7.817407 -0.568147  # second to last equiv to 1.534222
--
efp C6H6 -0.60075437  1.36443336  0.78647823  3.137879 1.557344 -2.568550
"""

ans5 = {
    "fragment_files": ["C6H6", "C6H6"],
    "hint_types": ["xyzabc", "xyzabc"],
    "geom_hints": [
        [-0.30448173, -2.24210052, -0.29383131, -0.642499, 7.817407, -0.568147],
        [-0.60075437, 1.36443336, 0.78647823, 3.137879, 1.557344, -2.568550],
    ],
    "geom": [],
    "elbl": [],
    "fragment_charges": [None],
    "fragment_multiplicities": [None],
    "fragment_separators": [],
}

fullans5b = {"efp": {}}
fullans5b["efp"]["hint_types"] = ans5["hint_types"]
fullans5b["efp"]["geom_hints"] = ans5["geom_hints"]
fullans5b["efp"]["units"] = "Bohr"
fullans5b["efp"]["fix_com"] = True
fullans5b["efp"]["fix_orientation"] = True
fullans5b["efp"]["fix_symmetry"] = "c1"
fullans5b["efp"]["fragment_files"] = ["c6h6", "c6h6"]


def test_psi4_efp_5a():
    subject = subject5

    hintsans = [
        [(val / qcelemental.constants.bohr2angstroms if i < 3 else val) for i, val in enumerate(ans5["geom_hints"][0])],
        [(val / qcelemental.constants.bohr2angstroms if i < 3 else val) for i, val in enumerate(ans5["geom_hints"][1])],
    ]
    hintsans[0][4] = 1.534222
    fullans = copy.deepcopy(fullans5b)
    fullans["efp"]["units"] = "Angstrom"
    fullans["efp"]["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans5, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": final efp")

    hintsstd = qcelemental.util.standardize_efp_angles_units("Angstrom", final["efp"]["geom_hints"])
    final["efp"]["geom_hints"] = hintsstd
    fullans["efp"]["geom_hints"] = hintsans
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": final efp standardized")


def test_psi4_efp_5b():
    subject = subject5 + "\nunits bohr"

    ans = copy.deepcopy(ans5)
    ans["units"] = "Bohr"
    fullans = copy.deepcopy(fullans5b)
    fullans["efp"]["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": final efp")


def test_psi4_efp_5c():
    """fix_orientation not mol kw"""
    subject = subject5 + "\nno_com\nfix_orientation\nsymmetry c1"

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_psi4_efp_5d():
    subject = subject5 + "\nno_com\nno_reorient\nsymmetry c1\nunits a.u."

    ans = copy.deepcopy(ans5)
    ans["units"] = "Bohr"
    ans["fix_com"] = True
    ans["fix_orientation"] = True
    ans["fix_symmetry"] = "c1"
    fullans = copy.deepcopy(fullans5b)
    fullans["efp"]["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": final")


def test_psi4_efp_5e():
    """symmetry w/efp"""
    subject = subject5 + "symmetry cs\nunits a.u."

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_string(subject, return_processed=True)


subject6 = """
    0 1
    O1    0         0     0.118720
    h2   -0.753299, 0.0, -0.474880

    H3    0.753299, 0.0, -0.474880

    --
    efp h2O -2.12417561  1.22597097 -0.95332054 -2.902133 -4.5481863 -1.953647  # second to last equiv to 1.734999
 --
efp ammoniA
     0.98792    1.87681    2.85174
units au
     1.68798    1.18856    3.09517

     1.45873    2.55904    2.27226

"""

ans6 = {
    "units": "Bohr",
    "geom": [0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
    "elbl": ["O1", "h2", "H3"],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "fragment_separators": [],
    "fragment_files": ["h2O", "ammoniA"],
    "geom_hints": [
        [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
        [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
    ],
    "hint_types": ["xyzabc", "points"],
}

fullans6 = {
    "qm": {
        "geom": np.array([0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880]),
        "elea": np.array([16, 1, 1]),
        "elez": np.array([8, 1, 1]),
        "elem": np.array(["O", "H", "H"]),
        "mass": np.array([15.99491462, 1.00782503, 1.00782503]),
        "real": np.array([True, True, True]),
        "elbl": np.array(["1", "2", "3"]),
        "units": "Bohr",
        "fix_com": True,
        "fix_orientation": True,
        "fix_symmetry": "c1",
        "fragment_charges": [0.0],
        "fragment_multiplicities": [1],
        "fragment_separators": [],
        "molecular_charge": 0.0,
        "molecular_multiplicity": 1,
    },
    "efp": {
        "fragment_files": ["h2o", "ammonia"],
        "geom_hints": [
            [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
            [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
        ],
        "hint_types": ["xyzabc", "points"],
        "units": "Bohr",
        "fix_com": True,
        "fix_orientation": True,
        "fix_symmetry": "c1",
    },
}


def test_psi4_qmefp_6a():
    subject = subject6
    fullans = copy.deepcopy(fullans6)
    fullans["qm"]["provenance"] = _string_prov_stamp
    fullans["efp"]["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans6, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")

    hintsstd = qcelemental.util.standardize_efp_angles_units("Bohr", final["efp"]["geom_hints"])
    final["efp"]["geom_hints"] = hintsstd
    fullans["efp"]["geom_hints"][0][4] = 1.734999
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": final efp standardized")


def test_psi4_qmefp_6b():
    subject = subject6.replace("au", "ang")

    ans = copy.deepcopy(ans6)
    ans["units"] = "Angstrom"

    fullans = copy.deepcopy(fullans6)
    fullans["qm"]["units"] = "Angstrom"
    fullans["efp"]["units"] = "Angstrom"
    fullans["qm"]["provenance"] = _string_prov_stamp
    fullans["efp"]["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")


def test_psi4_qmefpformat_error_6c():
    """try to give chgmult to an efp"""

    subject = subject6.replace("    efp h2O", "0 1\n    efp h2O")

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True)


def test_qmefp_array_6d():

    fullans = copy.deepcopy(fullans6)
    fullans["qm"]["provenance"] = _arrays_prov_stamp
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(
        units="Bohr",
        geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
        elbl=["O1", "h2", "H3"],
        fragment_files=["h2O", "ammoniA"],
        geom_hints=[
            [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
            [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
        ],
        hint_types=["xyzabc", "points"],
    )

    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")


def test_qmefp_badhint_error_6e():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "efp1"],
        )

    assert "hint_types not among 'xyzabc', 'points', 'rotmat'" in str(e.value)


def test_qmefp_badefpgeom_error_6f():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, None, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "points"],
        )

    assert "Un float-able elements in geom_hints" in str(e.value)


def test_qmefp_badhintgeom_error_6g():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["points", "xyzabc"],
        )

    assert "EFP hint type points not 9 elements" in str(e.value)


def test_qmefp_badfragfile_error_6h():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", 5],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "points"],
        )

    assert "fragment_files not strings" in str(e.value)


def test_qmefp_hintlen_error_6i():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "points", "points"],
        )

    assert "Missing or inconsistent length among efp quantities" in str(e.value)


def test_qmefp_fixcom_error_6j():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            fix_com=False,
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "points"],
        )

    assert "Invalid fix_com (False) with extern (True)" in str(e.value)


def test_qmefp_fixori_error_6k():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_input_arrays(
            units="Bohr",
            fix_orientation=False,
            geom=[0.0, 0.0, 0.118720, -0.753299, 0.0, -0.474880, 0.753299, 0.0, -0.474880],
            elbl=["O1", "h2", "H3"],
            fragment_files=["h2O", "ammoniA"],
            geom_hints=[
                [-2.12417561, 1.22597097, -0.95332054, -2.902133, -4.5481863, -1.953647],
                [0.98792, 1.87681, 2.85174, 1.68798, 1.18856, 3.09517, 1.45873, 2.55904, 2.27226],
            ],
            hint_types=["xyzabc", "points"],
        )

    assert "Invalid fix_orientation (False) with extern (True)" in str(e.value)


# QCEL@using_pylibefp
# QCELdef test_psi4_qmefp_6d():
# QCEL    subject = subject6
# QCEL
# QCEL    fullans = copy.deepcopy(fullans6)
# QCEL    fullans['efp']['geom'] = np.array([-2.22978429,  1.19270015, -0.99721732, -1.85344873,  1.5734809 ,
# QCEL        0.69660583, -0.71881655,  1.40649303, -1.90657336,  0.98792   ,
# QCEL        1.87681   ,  2.85174   ,  2.31084386,  0.57620385,  3.31175679,
# QCEL        1.87761143,  3.16604791,  1.75667803,  0.55253064,  2.78087794,
# QCEL        4.47837555])
# QCEL    fullans['efp']['elea'] = np.array([16, 1, 1, 14, 1, 1, 1])
# QCEL    fullans['efp']['elez'] = np.array([8, 1, 1, 7, 1, 1, 1])
# QCEL    fullans['efp']['elem'] = np.array(['O', 'H', 'H', 'N', 'H', 'H', 'H'])
# QCEL    fullans['efp']['mass'] = np.array([15.99491462, 1.00782503, 1.00782503, 14.00307400478, 1.00782503, 1.00782503, 1.00782503])
# QCEL    fullans['efp']['real'] = np.array([True, True, True, True, True, True, True])
# QCEL    fullans['efp']['elbl'] = np.array(['_a01o1', '_a02h2', '_a03h3', '_a01n1', '_a02h2', '_a03h3', '_a04h4'])
# QCEL    fullans['efp']['fragment_separators'] = [3]
# QCEL    fullans['efp']['fragment_charges'] = [0., 0.]
# QCEL    fullans['efp']['fragment_multiplicities'] = [1, 1]
# QCEL    fullans['efp']['molecular_charge'] = 0.
# QCEL    fullans['efp']['molecular_multiplicity'] = 1
# QCEL    fullans['efp']['hint_types'] = ['xyzabc', 'xyzabc']
# QCEL    fullans['efp']['geom_hints'][1] = [1.093116487139866, 1.9296501432128303, 2.9104336205167156, -1.1053108079381473, 2.0333070957565544, -1.488586877218809]
# QCEL
# QCEL    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True)
# QCEL
# QCEL    import pylibefp
# QCEL    efpobj = pylibefp.from_dict(final['efp'])
# QCEL    efpfinal = efpobj.to_dict()
# QCEL    efpfinal = qcelemental.molparse.from_arrays(speclabel=False, domain='efp', **efpfinal)
# QCEL
# QCEL    assert compare_molrecs(fullans['qm'], final['qm'], tnm() + ': full qm')
# QCEL    assert compare_molrecs(fullans['efp'], efpfinal, tnm() + ': full efp')

subject7 = """\
5
   stuffs
6Li 0.0 0.0 0.0
H_specIAL@2.014101  100 0 0
@Ne 2 4 6
h .0,1,2
Gh(he3) 0 1 3
"""

ans7 = {
    "geom": [0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0],
    "elbl": ["6Li", "H_specIAL@2.014101", "@Ne", "h", "Gh(he3)"],
    "units": "Angstrom",
    "geom_hints": [],  # shouldn't be needed
}

fullans7 = {
    "geom": np.array([0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0]),
    "elea": np.array([6, 2, 20, 1, 4]),
    "elez": np.array([3, 1, 10, 1, 2]),
    "elem": np.array(["Li", "H", "Ne", "H", "He"]),
    "mass": np.array([6.015122794, 2.014101, 19.99244017542, 1.00782503, 4.00260325415]),
    "real": np.array([True, True, False, True, False]),
    "elbl": np.array(["", "_special", "", "", "3"]),
    "units": "Angstrom",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [2],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 2,
}


def test_xyzp_qm_7a():
    """XYZ doesn't fit into psi4 string"""
    subject = subject7

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True, dtype="psi4")


def test_xyzp_qm_7b():
    """XYZ doesn't fit into strict xyz string"""
    subject = subject7

    with pytest.raises(qcelemental.MoleculeFormatError):
        qcelemental.molparse.from_string(subject, return_processed=True, dtype="xyz")


def test_xyzp_qm_7c():
    subject = subject7
    fullans = copy.deepcopy(fullans7)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, dtype="xyz+")
    assert compare_recursive(ans7, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full qm")


def test_xyzp_qm_7d():
    subject = subject7.replace("5", "5 au ")
    subject = subject.replace("stuff", "-1 3 slkdjfl2 32#$^& ")

    ans = copy.deepcopy(ans7)
    ans["units"] = "Bohr"
    ans["molecular_charge"] = -1.0
    ans["molecular_multiplicity"] = 3

    fullans = copy.deepcopy(fullans7)
    fullans["units"] = "Bohr"
    fullans["fragment_charges"] = [-1.0]
    fullans["fragment_multiplicities"] = [3]
    fullans["molecular_charge"] = -1.0
    fullans["molecular_multiplicity"] = 3
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, dtype="xyz+")
    assert compare_recursive(ans, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full qm")


def test_xyzp_qm_7e():
    subject = subject7.replace("5", "5 ang")
    fullans = copy.deepcopy(fullans7)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, dtype="xyz+")
    assert compare_recursive(ans7, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full qm")


subject8 = """\
3
   stuffs
Li 0.0 0.0 0.0
1  100 0 0
Ne 2 4 6
h .0,1,2
 2 0 1 3
"""

ans8 = {
    "geom": [0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0],
    "elbl": ["Li", "1", "Ne", "h", "2"],
    "units": "Angstrom",
    "geom_hints": [],  # shouldn't be needed
}

fullans8 = {
    "geom": np.array([0.0, 0.0, 0.0, 100.0, 0.0, 0.0, 2.0, 4.0, 6.0, 0.0, 1.0, 2.0, 0.0, 1.0, 3.0]),
    "elea": np.array([7, 1, 20, 1, 4]),
    "elez": np.array([3, 1, 10, 1, 2]),
    "elem": np.array(["Li", "H", "Ne", "H", "He"]),
    "mass": np.array([7.016004548, 1.00782503, 19.99244017542, 1.00782503, 4.00260325415]),
    "real": np.array([True, True, True, True, True]),
    "elbl": np.array(["", "", "", "", ""]),
    "units": "Angstrom",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [2],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 2,
}


def test_xyzp_qm_8a():
    subject = subject8
    fullans = copy.deepcopy(fullans8)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, dtype="xyz+")
    assert compare_recursive(ans8, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full qm", atol=1.0e-4)


fullans10qm = {
    "geom": np.array([0.0, 0.0, 0.0]),
    "elea": np.array([12]),
    "elez": np.array([6]),
    "elem": np.array(["C"]),
    "mass": np.array([12.0]),
    "real": np.array([True]),
    "elbl": np.array([""]),
    "units": "Angstrom",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
}
fullans10efp = {
    "fragment_files": ["cl2"],
    "hint_types": ["xyzabc"],
    "geom_hints": [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0]],
    "units": "Angstrom",
    "fix_com": True,
    "fix_orientation": True,
    "fix_symmetry": "c1",
}
blankqm = {
    "geom": np.array([]),
    "elea": np.array([]),
    "elez": np.array([]),
    "elem": np.array([]),
    "mass": np.array([]),
    "real": np.array([]),
    "elbl": np.array([]),
    "units": "Angstrom",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
}
blankefp = {
    "fragment_files": [],
    "hint_types": [],
    "geom_hints": [],
    "units": "Angstrom",
    "fix_com": True,
    "fix_orientation": True,
    "fix_symmetry": "c1",
}


def test_arrays_10a():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": True,
    }

    fullans = {"qm": copy.deepcopy(fullans10qm), "efp": copy.deepcopy(fullans10efp)}
    fullans["qm"]["fix_com"] = True
    fullans["qm"]["fix_orientation"] = True
    fullans["qm"]["fix_symmetry"] = "c1"
    fullans["qm"]["provenance"] = _arrays_prov_stamp
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10b():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": False,
        "enable_efp": True,
    }

    fullans = {"efp": fullans10efp}
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10c():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": False,
    }

    fullans = {"qm": fullans10qm}
    fullans["qm"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10d():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_efp": "none",
    }

    fullans = {"qm": fullans10qm}
    fullans["qm"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10e():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_efp": "minimal",
    }

    fullans = {"qm": fullans10qm, "efp": blankefp}
    fullans["qm"]["provenance"] = _arrays_prov_stamp
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10f():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_efp": "error",
    }

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_input_arrays(**subject)


def test_arrays_10g():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": False,
        "enable_efp": True,
        "missing_enabled_return_efp": "none",
    }

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10h():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": False,
        "enable_efp": True,
        "missing_enabled_return_efp": "minimal",
    }

    fullans = {"efp": blankefp}
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10i():
    subject = {
        "geom": [0, 0, 0],
        "elem": ["C"],
        "enable_qm": False,
        "enable_efp": True,
        "missing_enabled_return_efp": "error",
    }

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_input_arrays(**subject)


def test_arrays_10j():
    subject = {"geom": [0, 0, 0], "elem": ["C"], "enable_qm": True, "enable_efp": False}

    fullans = {"qm": fullans10qm}
    fullans["qm"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10k():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_qm": "none",
    }

    fullans = {"efp": fullans10efp}
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10l():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_qm": "minimal",
    }

    fullans = {"qm": copy.deepcopy(blankqm), "efp": fullans10efp}
    fullans["qm"]["fix_com"] = True
    fullans["qm"]["fix_orientation"] = True
    fullans["qm"]["fix_symmetry"] = "c1"
    fullans["qm"]["provenance"] = _arrays_prov_stamp
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10m():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": True,
        "missing_enabled_return_qm": "error",
    }

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_input_arrays(**subject)


def test_arrays_10n():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": False,
        "enable_efp": True,
    }

    fullans = {"efp": fullans10efp}
    fullans["efp"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_arrays_10o():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": False,
        "missing_enabled_return_qm": "none",
    }

    final = qcelemental.molparse.from_input_arrays(**subject)
    with pytest.raises(KeyError):
        final["qm"]
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10p():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": False,
        "missing_enabled_return_qm": "minimal",
    }

    fullans = {"qm": blankqm}
    fullans["qm"]["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_input_arrays(**subject)
    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    with pytest.raises(KeyError):
        final["efp"]


def test_arrays_10q():
    subject = {
        "fragment_files": ["cl2"],
        "hint_types": ["xyzabc"],
        "geom_hints": [[0, 0, 0, 0, 0, 0]],
        "enable_qm": True,
        "enable_efp": False,
        "missing_enabled_return_qm": "error",
    }

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_input_arrays(**subject)


def test_strings_10r():
    subject = ""

    final = qcelemental.molparse.from_string(
        subject, enable_qm=True, enable_efp=True, missing_enabled_return_qm="none", missing_enabled_return_efp="none"
    )

    print("final", final)
    with pytest.raises(KeyError):
        final["qm"]
    with pytest.raises(KeyError):
        final["efp"]


def test_strings_10s():
    subject = ""

    final = qcelemental.molparse.from_string(
        subject,
        enable_qm=True,
        enable_efp=True,
        missing_enabled_return_qm="minimal",
        missing_enabled_return_efp="minimal",
    )

    fullans = {"qm": blankqm, "efp": blankefp}
    fullans["qm"]["provenance"] = _string_prov_stamp
    fullans["efp"]["provenance"] = _string_prov_stamp

    assert compare_molrecs(fullans["qm"], final["qm"], tnm() + ": full qm")
    assert compare_molrecs(fullans["efp"], final["efp"], tnm() + ": full efp")


def test_strings_10t():
    subject = ""

    with pytest.raises(qcelemental.ValidationError):
        qcelemental.molparse.from_string(
            subject,
            enable_qm=True,
            enable_efp=True,
            missing_enabled_return_qm="error",
            missing_enabled_return_efp="error",
        )


def test_qmol_11c():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""nocom\n8 0 0 0\n1 1 0 0""", dtype="psi4")
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11d():
    fullans = copy.deepcopy(fullans1a)
    fullans.update({"variables": [], "geom_unsettled": [["0", "0", "0"], ["1", "0", "0"]]})
    fullans.pop("geom")
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""nocom\n8 0 0 0\n1 1 0 0""", dtype="psi4+")
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11e():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""2\n\nO 0 0 0 \n1 1 0 0 """, dtype="xyz", fix_com=True)
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11g():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _arrays_prov_stamp

    asdf = qcelemental.molparse.from_arrays(geom=[0.0, 0.0, 0.0, 1.0, 0.0, 0.0], elez=[8, 1], fix_com=True, verbose=2)
    assert compare_molrecs(fullans, asdf, 4)


def test_qmol_11h():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""nocom\n8 0 0 0\n1 1 0 0""")
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11i():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""nocom\n8 0 0 0\n1 1 0 0""")
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11j():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _string_prov_stamp

    asdf = qcelemental.molparse.from_string("""2\n\nO 0 0 0 \n1 1 0 0 """, fix_com=True)
    assert compare_molrecs(fullans, asdf["qm"], 4)


def test_qmol_11p():
    fullans = copy.deepcopy(fullans1a)
    fullans["provenance"] = _arrays_prov_stamp

    asdf = qcelemental.molparse.from_arrays(
        geom=[0.0, 0.0, 0.0, 1.0, 0.0, 0.0], elez=[8, 1], fix_com=True, units="AngSTRom"
    )
    assert compare_molrecs(fullans, asdf, 4)


def test_qmol_11q():
    with pytest.raises(KeyError):
        qcelemental.molparse.from_string("""2\n\nO 0 0 0 \n1 1 0 0 """, fix_com=True, dtype="psi3")


# QCELdef test_qmol_12():
# QCEL    asdf = qcdb.Molecule(geom=[ 0.,  0.,  0.,  1.,  0.,  0.], elez=[8, 1], fix_com=True)
# QCEL    assess_mol_11(asdf, 'qcdb.Molecule(geom, elez)')
# QCEL
# QCEL    import json
# QCEL    smol = json.dumps(asdf.to_dict(np_out=False))
# QCEL    dmol = json.loads(smol)
# QCEL
# QCEL    asdf2 = qcdb.Molecule(dmol)
# QCEL    assess_mol_11(asdf, 'qcdb.Molecule(jsondict)')

subject12 = """
 0 1
 1
 8 1 0.95
 O 2 1.40 1 A
 H 3 0.95 2 A 1 120.0

A = 105.0
"""

fullans12 = {
    "elbl": np.array(["", "", "", ""]),
    "elea": np.array([1, 16, 16, 1]),
    "elem": np.array(["H", "O", "O", "H"]),
    "elez": np.array([1, 8, 8, 1]),
    "fix_com": False,
    "fix_orientation": False,
    "fragment_charges": [0.0],
    "fragment_multiplicities": [1],
    "fragment_separators": [],
    "geom_unsettled": [[], ["1", "0.95"], ["2", "1.40", "1", "A"], ["3", "0.95", "2", "A", "1", "120.0"]],
    "mass": np.array([1.00782503, 15.99491462, 15.99491462, 1.00782503]),
    "molecular_charge": 0.0,
    "molecular_multiplicity": 1,
    "real": np.array([True, True, True, True]),
    "units": "Angstrom",
    "variables": [["A", 105.0]],
    "fix_symmetry": "c1",
}
ans12 = {
    "elbl": ["1", "8", "O", "H"],
    "fragment_charges": [0.0],
    "fragment_files": [],
    "fragment_multiplicities": [1],
    "fragment_separators": [],
    "geom_hints": [],
    "geom_unsettled": [[], ["1", "0.95"], ["2", "1.40", "1", "A"], ["3", "0.95", "2", "A", "1", "120.0"]],
    "hint_types": [],
    "variables": [("A", "105.0")],
    "fix_symmetry": "c1",
}


def test_psi4_qm_12a():
    subject = subject12
    fullans = copy.deepcopy(fullans12)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, fix_symmetry="c1")
    assert compare_recursive(ans12, intermed, tnm() + ": intermediate")
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_tooclose_error():
    subject = """2 -1 -1 -1\n2 -1 -1 -1.05"""

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_string(subject)

    assert "too close" in str(e.value)


def test_cartbeforezmat_error():
    subject = """He 0 0 0\nHe 1 2"""

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_string(subject)

    assert "Mixing Cartesian and Zmat" in str(e.value)


def test_jumbledzmat_error():
    subject = """He\nHe 1 2. 2 100. 3 35.\nHe 1 2."""

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_string(subject)

    assert "aim for lower triangular" in str(e.value)


def test_steepzmat_error():
    subject = """He\nHe 1 2.\nHe 1 2. 2 100. 3 35."""

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_string(subject)

    assert "aim for lower triangular" in str(e.value)


def test_zmatvar_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz", elem=["Rn", "Rn"], variables=[["bond", 2.0, "badextra"]], geom_unsettled=[[], ["1", "bond"]]
        )

    assert "Variables should come in pairs" in str(e.value)


def test_toomanyfrag_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["ar1", "42AR2"],
            fragment_multiplicities=[3, 3],
            fragment_separators=[1, 2],
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "zero-length fragment" in str(e.value)


def test_fragsep_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["ar1", "42AR2"],
            fragment_multiplicities=[3, 3],
            fragment_separators=np.array(["1"]),
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "unable to perform trial np.split on geometry" in str(e.value)


def test_cartzmatcart():
    subject = """1 1
# This part is just a normal Cartesian geometry specification for benzene
C          0.710500000000    -0.794637665924    -1.230622098778
C          1.421000000000    -0.794637665924     0.000000000000
C          0.710500000000    -0.794637665924     1.230622098778
C         -0.710500000000    -0.794637665924     1.230622098778
H          1.254500000000    -0.794637665924    -2.172857738095
H         -1.254500000000    -0.794637665924     2.172857738095
C         -0.710500000000    -0.794637665924    -1.230622098778
C         -1.421000000000    -0.794637665924     0.000000000000
H          2.509000000000    -0.794637665924     0.000000000000
H          1.254500000000    -0.794637665924     2.172857738095
H         -1.254500000000    -0.794637665924    -2.172857738095
H         -2.509000000000    -0.794637665924     0.000000000000
# And the hydronium part is specified using a zmatrix, referencing the benzene coordinates
X  1  CC  3  30   2  A2
O  13 R   1  90   2  90
H  14 OH  13 TDA  1  0
H  14 OH  15 TDA  13 A1
H  14 OH  15 TDA  13 -A1
He         3.710500000000    -0.794637665924    -1.230622098778

CC    = 1.421
CH    = 1.088
A1    = 120.0
A2    = 180.0
OH    = 1.05
R     = 4.0
"""

    qcelemental.molparse.from_string(subject)


def test_fixcom_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(elez=[3], molecular_charge=1, geom=[0, 0, 0], fix_com="thanks!")

    assert "Invalid fix_com" in str(e.value)


def test_fixori_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(elez=[3], molecular_charge=1, geom=[0, 0, 0], fix_orientation=-1)

    assert "Invalid fix_orientation" in str(e.value)


def test_units_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(elez=[3], molecular_charge=1, geom=[0, 0, 0], units="furlong")

    assert "Invalid molecule geometry units" in str(e.value)


def test_domain_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(domain="kitten")

    assert "Topology domain kitten not available" in str(e.value)


def test_natom_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(elem=["C"], elea=[12, 13], geom=[0, 0, 0])

    assert "Dimension mismatch natom" in str(e.value)


def test_incompletefrag_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["ar1", "42AR2"],
            fragment_multiplicities=[3, 3],
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "Fragment quantities given without separation info" in str(e.value)


def test_badmult_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["ar1", "42AR2"],
            fragment_multiplicities=[-3, 3],
            fragment_separators=np.array([1]),
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "fragment_multiplicities not among None or positive integer" in str(e.value)


def test_badchg_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["ar1", "42AR2"],
            fragment_charges=[[], {}],
            fragment_separators=np.array([1]),
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "fragment_charges not among None or float" in str(e.value)


def test_fraglen_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            domain="qmvz",
            speclabel=True,
            elbl=["na", "cl"],
            fragment_charges=[1, -1, 0],
            fragment_separators=np.array([1]),
            geom_unsettled=[[], ["1", "bond"]],
            hint_types=[],
            units="Bohr",
            variables=[("bond", "3")],
        )

    assert "mismatch among fragment quantities" in str(e.value)


def test_zmatfragarr_14a():
    fullans = copy.deepcopy(fullans14)
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        domain="qmvz",
        speclabel=True,
        elbl=["ar1", "42AR2"],
        fragment_multiplicities=[3, 3],
        fragment_separators=[1],
        geom_unsettled=[[], ["1", "bond"]],
        hint_types=[],
        units="Bohr",
        variables=[("bond", "3")],
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")


def test_zmatfragarr_14b():
    fullans = copy.deepcopy(fullans14)
    fullans["elbl"] = ["", ""]
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        domain="qmvz",
        speclabel=False,
        elez=[18, 18],
        elea=np.array([40, 42]),
        real=[True, True],
        fragment_multiplicities=[3, 3],
        fragment_separators=[1],
        geom_unsettled=[[], ["1", "bond"]],
        hint_types=[],
        units="Bohr",
        variables=[("bond", "3")],
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")


def test_zmatfragarr_14c():
    fullans = copy.deepcopy(fullans14)
    fullans["elbl"] = ["", ""]
    fullans["fix_com"] = True
    fullans["fix_orientation"] = True
    fullans["mass"] = fullans["mass"].tolist()  # other np vs. list diffs are hidden by compare_molrecs
    fullans["real"] = fullans["real"].tolist()
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        np_out=False,
        domain="qmvz",
        speclabel=False,
        elez=[18, 18],
        elea=[40, None],
        mass=[None, 41.96304574],
        real=[True, True],
        fragment_multiplicities=[3, 3],
        fragment_separators=[1],
        geom_unsettled=[[], ["1", "bond"]],
        hint_types=[],
        units="Bohr",
        fix_com=True,
        fix_orientation=True,
        variables=[("bond", "3")],
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")


subject14 = """
        0 3
        ar1
        --
        0 3
        42AR2 1 bond
    units a.u.
        bond =3.0
        """

fullans14 = {
    "elbl": np.array(["1", "2"]),
    "elea": np.array([40, 42]),
    "elem": np.array(["Ar", "Ar"]),
    "elez": np.array([18, 18]),
    "fix_com": False,
    "fix_orientation": False,
    "fragment_charges": [0.0, 0.0],
    "fragment_multiplicities": [3, 3],
    "fragment_separators": [1],
    "geom_unsettled": [[], ["1", "bond"]],
    "mass": np.array([39.96238312, 41.96304574]),
    "molecular_charge": 0.0,
    "molecular_multiplicity": 5,
    "real": np.array([True, True]),
    "units": "Bohr",
    "variables": [["bond", 3.0]],
}


def test_zmatfragstr_14d():
    subject = subject14
    fullans = copy.deepcopy(fullans14)
    fullans["provenance"] = _string_prov_stamp

    final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, verbose=2)
    assert compare_molrecs(fullans, final["qm"], tnm() + ": full")


def test_badprov0_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(geom=[1, 2, 3], elez=[4], provenance="mine")

    assert "Provenance entry is not dictionary" in str(e.value)


def test_badprov1_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=[1, 2, 3], elez=[4], provenance={"creator": ("psi", "tuple"), "routine": "buggy", "version": "0.1b"}
        )

    assert """Provenance key 'creator' should be string of creating program's name:""" in str(e.value)


def test_badprov2_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=[1, 2, 3], elez=[4], provenance={"creator": "", "routine": "buggy", "version": "my.vanity.version.13"}
        )

    assert """Provenance key 'version' should be a valid PEP 440 string:""" in str(e.value)


def test_badprov3_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=[1, 2, 3], elez=[4], provenance={"creator": "", "routine": 5, "version": "0.1b"}
        )

    assert """Provenance key 'routine' should be string of creating function's name:""" in str(e.value)


def test_badprov4_error():
    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=[1, 2, 3], elez=[4], provenance={"creators": "", "routine": "buggy", "version": "0.1b"}
        )

    assert """Provenance keys (['creator', 'routine', 'version']) incorrect:""" in str(e.value)


fullans17 = {
    "geom": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]),
    "elea": np.array([1, 32, 1]),
    "elez": np.array([1, 16, 1]),
    "elem": np.array(["H", "S", "H"]),
    "mass": np.array([1.00782503, 31.9720711744, 1.00782503]),
    "real": np.array([False, True, True]),
    "elbl": np.array(["", "", ""]),
    "units": "Bohr",
    "fix_com": False,
    "fix_orientation": False,
    "fragment_separators": [],
    "fragment_charges": [-1.0],
    "fragment_multiplicities": [1],
    "molecular_charge": -1.0,
    "molecular_multiplicity": 1,
    "connectivity": [(0, 1, 1.0), (1, 2, 1.0)],
}


def test_connectivity_17a():
    fullans = copy.deepcopy(fullans17)
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        geom=np.arange(9),
        units="Bohr",
        elez=[1, 16, 1],
        molecular_charge=-1,
        real=[False, True, True],
        connectivity=[(0, 1, 1), (1, 2, 1)],
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")


def test_connectivity_17b():
    fullans = copy.deepcopy(fullans17)
    fullans["provenance"] = _arrays_prov_stamp

    final = qcelemental.molparse.from_arrays(
        geom=np.arange(9),
        units="Bohr",
        elez=[1, 16, 1],
        molecular_charge=-1,
        real=[False, True, True],
        connectivity=[(2.0, 1, 1), (1, 0, 1)],
    )

    assert compare_molrecs(fullans, final, tnm() + ": full")


def test_connectivity_atindex_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=np.arange(9),
            units="Bohr",
            elez=[1, 16, 1],
            molecular_charge=-1,
            real=[False, True, True],
            connectivity=[(2.1, 1, 1), (1, 0, 1)],
        )

    assert "Connectivity first atom should be int" in str(e.value)


def test_connectivity_atrange_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=np.arange(9),
            units="Bohr",
            elez=[1, 16, 1],
            molecular_charge=-1,
            real=[False, True, True],
            connectivity=[(2, 1, 1), (1, -1, 1)],
        )

    assert "Connectivity second atom should be int" in str(e.value)


def test_connectivity_bondorder_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=np.arange(9),
            units="Bohr",
            elez=[1, 16, 1],
            molecular_charge=-1,
            real=[False, True, True],
            connectivity=[(2, 1, 1), (1, 0, 6)],
        )

    assert "Connectivity bond order should be float" in str(e.value)


def test_connectivity_type_error():

    with pytest.raises(qcelemental.ValidationError) as e:
        qcelemental.molparse.from_arrays(
            geom=np.arange(9),
            units="Bohr",
            elez=[1, 16, 1],
            molecular_charge=-1,
            real=[False, True, True],
            connectivity="wire",
        )

    assert "Connectivity entry is not of form" in str(e.value)


#'geom_unsettled': [[], ['1', '2.'], ['1', '2.', '2', '100.', '3', '35.']],

# final, intermed = qcelemental.molparse.from_string(subject, return_processed=True, verbose=2)
# import pprint
# pprint.pprint(final)
# pprint.pprint(intermed)

# assert False



import numpy as np
import pytest

import qcelemental as qcel
from qcelemental.models import basis

center_data = {
    "bs_sto3g_h": {
        "electron_shells": [
            {
                "harmonic_type": "spherical",
                "angular_momentum": [0],
                "exponents": [3.42525091, 0.62391373, 0.16885540],
                "coefficients": [[0.15432897, 0.53532814, 0.44463454]],
            }
        ]
    },
    "bs_sto3g_o": {
        "electron_shells": [
            {
                "harmonic_type": "spherical",
                "angular_momentum": [0],
                "exponents": [130.70939, 23.808861, 6.4436089],
                "coefficients": [[0.15432899, 0.53532814, 0.44463454]],
            },
            {
                "harmonic_type": "cartesian",
                "angular_momentum": [0, 1],
                "exponents": [5.0331513, 1.1695961, 0.3803890],
                "coefficients": [[-0.09996723, 0.39951283, 0.70011547], [0.15591629, 0.60768379, 0.39195739]],
            },
            {
                "harmonic_type": "cartesian",
                "angular_momentum": [0],
                "exponents": [5.0331513, 1.1695961, 0.3803890],
                "coefficients": [[-5.09996723, 0.39951283, 0.70011547], [0.15591629, 0.60768379, 0.39195739]],
            },
        ]
    },
    "bs_def2tzvp_zr": {
        "electron_shells": [
            {
                "harmonic_type": "spherical",
                "angular_momentum": [0],
                "exponents": [11.000000000, 9.5000000000, 3.6383667759, 0.76822026698],
                "coefficients": [
                    [-0.19075595259, 0.33895588754, 0.0000000, 0.0000000],
                    [0.0000000, 0.0000000, 1.0000000000, 0.0000000],
                ],
            },
            {
                "harmonic_type": "spherical",
                "angular_momentum": [2],
                "exponents": [4.5567957795, 1.2904939799, 0.51646987229],
                "coefficients": [
                    [-0.96190569023e-09, 0.20569990155, 0.41831381851],
                    [0.0000000, 0.0000000, 0.0000000],
                    [0.0000000, 0.0000000, 0.0000000],
                ],
            },
            {
                "harmonic_type": "spherical",
                "angular_momentum": [3],
                "exponents": [0.3926100],
                "coefficients": [[1.0000000]],
            },
        ],
        "ecp_electrons": 28,
        "ecp_potentials": [
            {
                "ecp_type": "scalar",
                "angular_momentum": [0],
                "r_exponents": [2, 2, 2, 2],
                "gaussian_exponents": [7.4880494, 3.7440249, 6.5842120, 3.2921060],
                "coefficients": [[135.15384419, 15.55244130, 19.12219811, 2.43637549]],
            },
            {
                "ecp_type": "spinorbit",
                "angular_momentum": [1],
                "r_exponents": [2, 2, 2, 2],
                "gaussian_exponents": [6.4453779, 3.2226886, 6.5842120, 3.2921060],
                "coefficients": [[87.78499169, 11.56406599, 19.12219811, 2.43637549]],
            },
        ],
    },
}


@pytest.fixture(scope="function")
def result_data_fixture():
    mol = qcel.models.Molecule.from_data(
        """
        O 0 0 0
        H 0 0 2
        H 0 2 0
    """
    )

    return {
        "molecule": mol,
        "driver": "energy",
        "model": {"method": "UFF"},
        "return_result": 5,
        "success": True,
        "properties": {},
        "provenance": {"creator": "qcel"},
        "stdout": "I ran.",
    }


@pytest.fixture(scope="function")
def wavefunction_data_fixture(result_data_fixture):
    bas = basis.BasisSet(
        name="custom_basis", center_data=center_data, atom_map=["bs_sto3g_o", "bs_sto3g_h", "bs_sto3g_h"]
    )
    c_matrix = np.random.rand(bas.nbf, bas.nbf)
    result_data_fixture["protocols"] = {"wavefunction": "all"}
    result_data_fixture["wavefunction"] = {
        "basis": bas,
        "restricted": True,
        "scf_orbitals_a": c_matrix,
        "orbitals_a": "scf_orbitals_a",
    }

    return result_data_fixture


@pytest.fixture(scope="function")
def optimization_data_fixture(result_data_fixture):

    trajectory = []
    energies = []
    for x in range(5):
        result = result_data_fixture.copy()
        result["return_result"] = x
        trajectory.append(result)
        energies.append(x)

    ret = {
        "initial_molecule": result_data_fixture["molecule"],
        "final_molecule": result_data_fixture["molecule"],
        "trajectory": trajectory,
        "energies": energies,
        "success": True,
        "provenance": {"creator": "qcel"},
        "input_specification": {"model": {"method": "UFF"}},
    }

    return ret


@pytest.mark.parametrize("center_name", center_data.keys())
def test_basis_shell_centers(center_name):
    assert basis.BasisCenter(**center_data[center_name])


def test_basis_set_build():
    bas = basis.BasisSet(
        name="custom_basis",
        center_data=center_data,
        atom_map=["bs_sto3g_o", "bs_sto3g_h", "bs_sto3g_h", "bs_def2tzvp_zr"],
    )

    assert len(bas.center_data) == 3
    assert len(bas.atom_map) == 4
    assert bas.nbf == 21

    es = bas.center_data["bs_sto3g_o"].electron_shells
    assert es[0].is_contracted() is False
    assert es[1].is_contracted() is False
    assert es[2].is_contracted()


def test_basis_electron_center_raises():
    data = center_data["bs_sto3g_h"]["electron_shells"][0].copy()

    # Check bad coefficient length
    bad_coef = data.copy()
    bad_coef["coefficients"] = [[5, 3]]
    with pytest.raises(ValueError) as e:
        basis.ElectronShell(**bad_coef)

    assert "does not match the" in str(e.value)

    # Check bad fused shell
    bad_fused = data.copy()
    bad_fused["angular_momentum"] = [0, 1]
    with pytest.raises(ValueError) as e:
        basis.ElectronShell(**bad_fused)

    assert "fused shell" in str(e.value)


def test_basis_ecp_center_raises():
    # Check coefficients
    data = center_data["bs_def2tzvp_zr"]["ecp_potentials"][0].copy()
    data["coefficients"] = [[5, 3]]

    with pytest.raises(ValueError):
        basis.ECPPotential(**data)

    # Check gaussian_exponents
    data = center_data["bs_def2tzvp_zr"]["ecp_potentials"][0].copy()
    data["gaussian_exponents"] = [5, 3]

    with pytest.raises(ValueError):
        basis.ECPPotential(**data)


def test_basis_map_raises():

    with pytest.raises(ValueError) as e:
        assert basis.BasisSet(name="custom_basis", center_data=center_data, atom_map=["something_odd"])


def test_result_build(result_data_fixture):
    ret = qcel.models.AtomicResult(**result_data_fixture)
    assert ret.wavefunction is None


def test_result_build_wavefunction_delete(wavefunction_data_fixture):
    del wavefunction_data_fixture["protocols"]
    ret = qcel.models.AtomicResult(**wavefunction_data_fixture)
    assert ret.wavefunction is None


def test_wavefunction_build(wavefunction_data_fixture):
    assert qcel.models.AtomicResult(**wavefunction_data_fixture)


def test_wavefunction_matrix_size_error(wavefunction_data_fixture):

    wavefunction_data_fixture["wavefunction"]["scf_orbitals_a"] = np.random.rand(2, 2)
    with pytest.raises(ValueError) as e:
        qcel.models.AtomicResult(**wavefunction_data_fixture)

    assert "castable to shape" in str(e.value)


def test_wavefunction_return_result_pointer(wavefunction_data_fixture):

    del wavefunction_data_fixture["wavefunction"]["scf_orbitals_a"]
    with pytest.raises(ValueError) as e:
        qcel.models.AtomicResult(**wavefunction_data_fixture)

    assert "does not exist" in str(e.value)


@pytest.mark.parametrize(
    "protocol, restricted, provided, expected",
    [
        ("none", True, ["orbitals_a", "orbitals_b"], []),
        (None, True, ["orbitals_a", "orbitals_b"], []),
        ("all", False, ["orbitals_a", "orbitals_b"], ["orbitals_a", "orbitals_b"]),
        ("all", True, ["orbitals_a", "orbitals_b"], ["orbitals_a"]),
        (
            "orbitals_and_eigenvalues",
            False,
            ["orbitals_a", "orbitals_b", "fock_a", "fock_b"],
            ["orbitals_a", "orbitals_b"],
        ),
        (
            "orbitals_and_eigenvalues",
            True,
            ["orbitals_a", "orbitals_b", "eigenvalues_a", "fock_a", "fock_b"],
            ["orbitals_a", "eigenvalues_a"],
        ),
        ("return_results", True, ["orbitals_a", "fock_a", "fock_b"], ["orbitals_a", "fock_a"]),
    ],
)
def test_wavefunction_protocols(protocol, restricted, provided, expected, wavefunction_data_fixture):

    wfn_data = wavefunction_data_fixture["wavefunction"]

    if protocol is None:
        wavefunction_data_fixture.pop("protocols")
    else:
        wavefunction_data_fixture["protocols"]["wavefunction"] = protocol

    wfn_data["restricted"] = restricted
    bas = wfn_data["basis"]

    for name in provided:
        scf_name = "scf_" + name
        wfn_data[name] = scf_name
        if "eigen" in name:
            wfn_data[scf_name] = np.random.rand(bas.nbf)
        else:
            wfn_data[scf_name] = np.random.rand(bas.nbf, bas.nbf)

    wfn = qcel.models.AtomicResult(**wavefunction_data_fixture)

    if len(expected) == 0:
        assert wfn.wavefunction is None
    else:
        expected_keys = set(expected) | {"scf_" + x for x in expected} | {"basis", "restricted"}
        assert wfn.wavefunction.dict().keys() == expected_keys


@pytest.mark.parametrize(
    "keep, indices",
    [(None, [0, 1, 2, 3, 4]), ("all", [0, 1, 2, 3, 4]), ("initial_and_final", [0, 4]), ("final", [4]), ("none", [])],
)
def test_optimization_trajectory_protocol(keep, indices, optimization_data_fixture):

    if keep is not None:
        optimization_data_fixture["protocols"] = {"trajectory": keep}
    opt = qcel.models.OptimizationResult(**optimization_data_fixture)

    assert len(opt.trajectory) == len(indices)
    for result, index in zip(opt.trajectory, indices):
        assert result.return_result == index


@pytest.mark.parametrize(
    "default, defined, default_result, defined_result",
    [(None, None, True, None), (False, {"a": True}, False, {"a": True})],
)
def test_error_correction_protocol(default, defined, default_result, defined_result, result_data_fixture):
    policy = {}
    if default is not None:
        policy["default_policy"] = default
    if defined is not None:
        policy["policies"] = defined
    result_data_fixture["protocols"] = {"error_correction": policy}
    res = qcel.models.AtomicResult(**result_data_fixture)

    assert res.protocols.error_correction.default_policy == default_result
    assert res.protocols.error_correction.policies == defined_result


def test_error_correction_logic():
    # Make sure we are permissive by default
    correction_policy = qcel.models.results.ErrorCorrectionProtocol()
    assert correction_policy.allows("a")

    # Add ability to turn off error correction
    correction_policy = qcel.models.results.ErrorCorrectionProtocol(policies={"a": False})
    correction_policy.policies["a"] = False
    assert not correction_policy.allows("a")

    # Try no execution by default
    correction_policy = qcel.models.results.ErrorCorrectionProtocol(default_policy=False)
    assert not correction_policy.allows("a")

    # Make sure it is still overridable
    correction_policy = qcel.models.results.ErrorCorrectionProtocol(default_policy=False, policies={"a": True})
    assert correction_policy.allows("a")


def test_result_build_stdout_delete(result_data_fixture):
    result_data_fixture["protocols"] = {"stdout": False}
    ret = qcel.models.AtomicResult(**result_data_fixture)
    assert ret.stdout is None


def test_result_build_stdout(result_data_fixture):
    ret = qcel.models.AtomicResult(**result_data_fixture)
    assert ret.stdout == "I ran."


def test_failed_operation(result_data_fixture):
    water = qcel.models.Molecule.from_data(
        """
        O 0 0 0
        H 0 0 2
        H 0 2 0
    """
    )

    failed = qcel.models.FailedOperation(
        extras={"garbage": water},
        input_data=result_data_fixture,
        error={"error_type": "expected_testing_error", "error_message": "If you see this, its all good"},
    )
    assert isinstance(failed.error, qcel.models.ComputeError)
    assert isinstance(failed.dict(), dict)
    failed_json = failed.json()
    assert isinstance(failed_json, str)
    assert "its all good" in failed_json


def test_result_properties_array():
    lquad = [1, 2, 3, 2, 4, 5, 3, 5, 6]

    obj = qcel.models.AtomicResultProperties(
        scf_one_electron_energy="-5.0", scf_dipole_moment=[1, 2, 3], scf_quadrupole_moment=lquad
    )

    assert pytest.approx(obj.scf_one_electron_energy) == -5.0
    assert obj.scf_dipole_moment.shape == (3,)
    assert obj.scf_quadrupole_moment.shape == (3, 3)

    assert obj.dict().keys() == {"scf_one_electron_energy", "scf_dipole_moment", "scf_quadrupole_moment"}
    assert np.array_equal(obj.scf_quadrupole_moment, np.array(lquad).reshape(3, 3))
    assert obj.dict()["scf_quadrupole_moment"] == lquad


@pytest.mark.parametrize(
    "smodel", ["molecule", "atomicresultproperties", "atomicinput", "atomicresult", "optimizationresult"]
)
def test_model_dictable(result_data_fixture, optimization_data_fixture, smodel):

    if smodel == "molecule":
        model = qcel.models.Molecule
        data = result_data_fixture["molecule"].dict()

    elif smodel == "atomicresultproperties":
        model = qcel.models.AtomicResultProperties
        data = {"scf_one_electron_energy": "-5.0", "scf_dipole_moment": [1, 2, 3], "ccsd_dipole_moment": None}

    elif smodel == "atomicinput":
        model = qcel.models.AtomicInput
        data = {k: result_data_fixture[k] for k in ["molecule", "model", "driver"]}

    elif smodel == "atomicresult":
        model = qcel.models.AtomicResult
        data = result_data_fixture

    elif smodel == "optimizationresult":
        model = qcel.models.OptimizationResult
        data = optimization_data_fixture

    instance = model(**data)
    assert model(**instance.dict())


def test_result_model_deprecations(result_data_fixture, optimization_data_fixture):

    with pytest.warns(DeprecationWarning):
        qcel.models.ResultProperties(scf_one_electron_energy="-5.0")

    with pytest.warns(DeprecationWarning):
        qcel.models.ResultInput(**{k: result_data_fixture[k] for k in ["molecule", "model", "driver"]})

    with pytest.warns(DeprecationWarning):
        qcel.models.Result(**result_data_fixture)

    with pytest.warns(DeprecationWarning):
        qcel.models.Optimization(**optimization_data_fixture)



from .chgmult import validate_and_fill_chgmult
from .from_arrays import from_arrays, from_input_arrays
from .from_schema import contiguize_from_fragment_pattern, from_schema
from .from_string import from_string
from .nucleus import parse_nucleus_label, reconcile_nucleus
from .to_schema import to_schema
from .to_string import to_string



# [Apr 2018] stolen directly from scipy so I can get an array back
#   https://github.com/scipy/scipy/blob/master/scipy/optimize/_hungarian.py
# [Apr 2019]
#   * apply MAINT: Switch np.where(c) for np.nonzero(c) 9757612
#   * transpose reduced_cost if cost_matrix had to be transposed

# Hungarian algorithm (Kuhn-Munkres) for solving the linear sum assignment
# problem. Taken from scikit-learn. Based on original code by Brian Clapper,
# adapted to NumPy by Gael Varoquaux.
# Further improvements by Ben Root, Vlad Niculae and Lars Buitinck.
#
# Copyright (c) 2008 Brian M. Clapper <bmc@clapper.org>, Gael Varoquaux
# Author: Brian M. Clapper, Gael Varoquaux
# License: 3-clause BSD

import numpy as np


def linear_sum_assignment(cost_matrix, return_cost=False):
    """Solve the linear sum assignment problem.

    The linear sum assignment problem is also known as minimum weight matching
    in bipartite graphs. A problem instance is described by a matrix C, where
    each C[i,j] is the cost of matching vertex i of the first partite set
    (a "worker") and vertex j of the second set (a "job"). The goal is to find
    a complete assignment of workers to jobs of minimal cost.

    Formally, let X be a boolean matrix where :math:`X[i,j] = 1` iff row i is
    assigned to column j. Then the optimal assignment has cost

    .. math::
        \\min \\sum_i \\sum_j C_{i,j} X_{i,j}

    s.t. each row is assignment to at most one column, and each column to at
    most one row.

    This function can also solve a generalization of the classic assignment
    problem where the cost matrix is rectangular. If it has more rows than
    columns, then not every row needs to be assigned to a column, and vice
    versa.

    The method used is the Hungarian algorithm, also known as the Munkres or
    Kuhn-Munkres algorithm.

    Parameters
    ----------
    cost_matrix : array
        The cost matrix of the bipartite graph.
    return_cost : bool, optional
        If True, also return a copy of the cost_matrix reduced to maximal
        zeros at the end of the Munkres algorithm.

    Returns
    -------
    row_ind, col_ind : array
        An array of row indices and one of corresponding column indices giving
        the optimal assignment. The cost of the assignment can be computed
        as ``cost_matrix[row_ind, col_ind].sum()``. The row indices will be
        sorted; in the case of a square cost matrix they will be equal to
        ``numpy.arange(cost_matrix.shape[0])``.
    (row_ind, col_ind), cost
        Only provided if `return_cost` is True.

    Notes
    -----
    .. versionadded:: 0.17.0

    Examples
    --------
    >>> cost = np.array([[4, 1, 3], [2, 0, 5], [3, 2, 2]])
    >>> from scipy.optimize import linear_sum_assignment
    >>> row_ind, col_ind = linear_sum_assignment(cost)
    >>> col_ind
    array([1, 0, 2])
    >>> cost[row_ind, col_ind].sum()
    5

    References
    ----------
    1. http://csclab.murraystate.edu/bob.pilgrim/445/munkres.html

    2. Harold W. Kuhn. The Hungarian Method for the assignment problem.
       *Naval Research Logistics Quarterly*, 2:83-97, 1955.

    3. Harold W. Kuhn. Variants of the Hungarian method for assignment
       problems. *Naval Research Logistics Quarterly*, 3: 253-258, 1956.

    4. Munkres, J. Algorithms for the Assignment and Transportation Problems.
       *J. SIAM*, 5(1):32-38, March, 1957.

    5. https://en.wikipedia.org/wiki/Hungarian_algorithm
    """
    cost_matrix = np.asarray(cost_matrix)
    if len(cost_matrix.shape) != 2:
        raise ValueError("expected a matrix (2-d array), got a %r array" % (cost_matrix.shape,))

    if not (np.issubdtype(cost_matrix.dtype, np.number) or cost_matrix.dtype == np.dtype(np.bool)):
        raise ValueError("expected a matrix containing numerical entries, got %s" % (cost_matrix.dtype,))

    if np.any(np.isinf(cost_matrix) | np.isnan(cost_matrix)):
        raise ValueError("matrix contains invalid numeric entries")

    if cost_matrix.dtype == np.dtype(np.bool):
        cost_matrix = cost_matrix.astype(np.int)

    # The algorithm expects more columns than rows in the cost matrix.
    if cost_matrix.shape[1] < cost_matrix.shape[0]:
        cost_matrix = cost_matrix.T
        transposed = True
    else:
        transposed = False

    state = _Hungary(cost_matrix)

    # No need to bother with assignments if one of the dimensions
    # of the cost matrix is zero-length.
    step = None if 0 in cost_matrix.shape else _step1

    while step is not None:
        step = step(state)

    if transposed:
        marked = state.marked.T
        reduced_cost = state.C.T
    else:
        marked = state.marked
        reduced_cost = state.C

    if return_cost:
        return np.nonzero(marked == 1), reduced_cost
    else:
        return np.nonzero(marked == 1)


class _Hungary(object):
    """State of the Hungarian algorithm.

    Parameters
    ----------
    cost_matrix : 2D matrix
        The cost matrix. Must have shape[1] >= shape[0].
    """

    def __init__(self, cost_matrix):
        self.C = cost_matrix.copy()

        n, m = self.C.shape
        self.row_uncovered = np.ones(n, dtype=bool)
        self.col_uncovered = np.ones(m, dtype=bool)
        self.Z0_r = 0
        self.Z0_c = 0
        self.path = np.zeros((n + m, 2), dtype=int)
        self.marked = np.zeros((n, m), dtype=int)

    def _clear_covers(self):
        """Clear all covered matrix cells"""
        self.row_uncovered[:] = True
        self.col_uncovered[:] = True


# Individual steps of the algorithm follow, as a state machine: they return
# the next step to be taken (function to be called), if any.


def _step1(state):
    """Steps 1 and 2 in the Wikipedia page."""

    # Step 1: For each row of the matrix, find the smallest element and
    # subtract it from every element in its row.
    state.C -= state.C.min(axis=1)[:, np.newaxis]
    # Step 2: Find a zero (Z) in the resulting matrix. If there is no
    # starred zero in its row or column, star Z. Repeat for each element
    # in the matrix.
    for i, j in zip(*np.nonzero(state.C == 0)):
        if state.col_uncovered[j] and state.row_uncovered[i]:
            state.marked[i, j] = 1
            state.col_uncovered[j] = False
            state.row_uncovered[i] = False

    state._clear_covers()
    return _step3


def _step3(state):
    """
    Cover each column containing a starred zero. If n columns are covered,
    the starred zeros describe a complete set of unique assignments.
    In this case, Go to DONE, otherwise, Go to Step 4.
    """
    marked = state.marked == 1
    state.col_uncovered[np.any(marked, axis=0)] = False

    if marked.sum() < state.C.shape[0]:
        return _step4


def _step4(state):
    """
    Find a noncovered zero and prime it. If there is no starred zero
    in the row containing this primed zero, Go to Step 5. Otherwise,
    cover this row and uncover the column containing the starred
    zero. Continue in this manner until there are no uncovered zeros
    left. Save the smallest uncovered value and Go to Step 6.
    """
    # We convert to int as numpy operations are faster on int
    C = (state.C == 0).astype(int)
    covered_C = C * state.row_uncovered[:, np.newaxis]
    covered_C *= np.asarray(state.col_uncovered, dtype=int)
    n = state.C.shape[0]
    m = state.C.shape[1]

    while True:
        # Find an uncovered zero
        row, col = np.unravel_index(np.argmax(covered_C), (n, m))
        if covered_C[row, col] == 0:
            return _step6
        else:
            state.marked[row, col] = 2
            # Find the first starred element in the row
            star_col = np.argmax(state.marked[row] == 1)
            if state.marked[row, star_col] != 1:
                # Could not find one
                state.Z0_r = row
                state.Z0_c = col
                return _step5
            else:
                col = star_col
                state.row_uncovered[row] = False
                state.col_uncovered[col] = True
                covered_C[:, col] = C[:, col] * (np.asarray(state.row_uncovered, dtype=int))
                covered_C[row] = 0


def _step5(state):
    """
    Construct a series of alternating primed and starred zeros as follows.
    Let Z0 represent the uncovered primed zero found in Step 4.
    Let Z1 denote the starred zero in the column of Z0 (if any).
    Let Z2 denote the primed zero in the row of Z1 (there will always be one).
    Continue until the series terminates at a primed zero that has no starred
    zero in its column. Unstar each starred zero of the series, star each
    primed zero of the series, erase all primes and uncover every line in the
    matrix. Return to Step 3
    """
    count = 0
    path = state.path
    path[count, 0] = state.Z0_r
    path[count, 1] = state.Z0_c

    while True:
        # Find the first starred element in the col defined by
        # the path.
        row = np.argmax(state.marked[:, path[count, 1]] == 1)
        if state.marked[row, path[count, 1]] != 1:
            # Could not find one
            break
        else:
            count += 1
            path[count, 0] = row
            path[count, 1] = path[count - 1, 1]

        # Find the first prime element in the row defined by the
        # first path step
        col = np.argmax(state.marked[path[count, 0]] == 2)
        if state.marked[row, col] != 2:
            col = -1
        count += 1
        path[count, 0] = path[count - 1, 0]
        path[count, 1] = col

    # Convert paths
    for i in range(count + 1):
        if state.marked[path[i, 0], path[i, 1]] == 1:
            state.marked[path[i, 0], path[i, 1]] = 0
        else:
            state.marked[path[i, 0], path[i, 1]] = 1

    state._clear_covers()
    # Erase all prime markings
    state.marked[state.marked == 2] = 0
    return _step3


def _step6(state):
    """
    Add the value found in Step 4 to every element of each covered row,
    and subtract it from every element of each uncovered column.
    Return to Step 4 without altering any stars, primes, or covered lines.
    """
    # the smallest uncovered value in the matrix
    if np.any(state.row_uncovered) and np.any(state.col_uncovered):
        minval = np.min(state.C[state.row_uncovered], axis=0)
        minval = np.min(minval[state.col_uncovered])
        state.C[~state.row_uncovered] += minval
        state.C[:, state.col_uncovered] -= minval
    return _step4



import pytest

from qcelemental.models import (
    AtomicInput,
    AtomicResult,
    AtomicResultProperties,
    ComputeError,
    FailedOperation,
    Molecule,
    Optimization,
    OptimizationInput,
    ProtoModel,
    Provenance,
)


def test_result_properties_default_skip():

    obj = AtomicResultProperties(scf_one_electron_energy="-5.0")

    assert pytest.approx(obj.scf_one_electron_energy) == -5.0

    assert obj.dict().keys() == {"scf_one_electron_energy"}


def test_result_properties_default_repr():
    obj = AtomicResultProperties(scf_one_electron_energy="-5.0")
    assert "none" not in str(obj).lower()
    assert "scf_one_electron_energy" in str(obj)
    assert len(str(obj)) < 100
    assert len(repr(obj)) < 100


def test_repr_provenance():

    prov = Provenance(creator="qcel", version="v0.3.2")

    assert "qcel" in str(prov)
    assert "qcel" in repr(prov)


def test_repr_compute_error():
    ce = ComputeError(error_type="random_error", error_message="this is bad")

    assert "random_error" in str(ce)
    assert "random_error" in repr(ce)


def test_repr_failed_op():
    fail_op = FailedOperation(error=ComputeError(error_type="random_error", error_message="this is bad"))
    assert (
        str(fail_op)
        == """FailedOperation(error=ComputeError(error_type='random_error', error_message='this is bad'))"""
    )


def test_repr_result():

    result = AtomicInput(
        **{"driver": "gradient", "model": {"method": "UFF"}, "molecule": {"symbols": ["He"], "geometry": [0, 0, 0]}}
    )
    assert "molecule_hash" in str(result)
    assert "molecule_hash" in repr(result)
    assert "'gradient'" in str(result)


def test_repr_optimization():

    opt = OptimizationInput(
        **{
            "input_specification": {"driver": "gradient", "model": {"method": "UFF"}},
            "initial_molecule": {"symbols": ["He"], "geometry": [0, 0, 0]},
        }
    )

    assert "molecule_hash" in str(opt)
    assert "molecule_hash" in repr(opt)


def test_model_custom_repr():
    class Model(ProtoModel):
        a: int

        def __repr__(self) -> str:
            return "Hello world!"

    m = Model(a=5)
    assert repr(m) == "Hello world!"
    assert "Model(" in str(m)

    class Model2(ProtoModel):
        a: int

        def __str__(self) -> str:
            return "Hello world!"

    m = Model2(a=5)
    assert "Model2(" in repr(m)
    assert str(m) == "Hello world!"



# remember to use re.VERBOSE with DECIMAL, NUCLEUS, NUMBER, VERSION_PATTERN

NUCLEUS = r"""(?:
   (?P<gh1>@)|(?P<gh2>Gh\())?                # optional ghost: @stuff or Gh(stuff) ...
        (                                    # mandatory element: AEuser or Zuser
            (?P<label1>
                (?P<A>\d+)?                  # optional mass number, A
                (?P<E>[A-Z]{1,3})            # mandatory atomic symbol, E
                (?P<user1>(_\w+)|(\d+))?) |  # optional user label: Enumber or E_alphanum
            (?P<label2>
                (?P<Z>\d{1,3})               # mandatory atomic number, Z
                (?P<user2>(_\w+))?)          # optional user label: Z_alphanum
        )
        (?:@(?P<mass>\d+\.\d+))?             # optional mass value [u]
   (?(gh2)\)                                 # ... ghost
          )"""

NUMBER = r"""(
    (?:[-+]?\d*\.\d+(?:[DdEe][-+]?\d+)?) |   # .num with optional sign, exponent, wholenum
    (?:[-+]?\d+\.\d*(?:[DdEe][-+]?\d+)?) |   # num. with optional sign, exponent, decimals
    (?:[-+]?\d+(?:[DdEe][-+]?\d+)?)          # num with optional sign, exponent
         )"""

DECIMAL = r"""(
    (?:[-+]?\d*\.\d+(?:[DdEe][-+]?\d+)?) |   # .num with optional sign, exponent, wholenum
    (?:[-+]?\d+\.\d*(?:[DdEe][-+]?\d+)?)     # num. with optional sign, exponent, decimals
         )"""

SEP = r"""[\t ,]+"""
ENDL = r"""[\t ,]*$"""

CHGMULT = r"""(?P<chg>""" + NUMBER + r")" + SEP + r"""(?P<mult>\d+)"""
CARTXYZ = r"(?P<x>" + NUMBER + r")" + SEP + r"(?P<y>" + NUMBER + r")" + SEP + r"(?P<z>" + NUMBER + r")"

# PEP 440 valid versions
# * to avoid dependency on module "packaging", copied from https://github.com/pypa/packaging/blob/master/packaging/version.py#L182-L213
# * Deliberately not anchored to the start and end of the string, to make it easier for 3rd party code to reuse
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
